use std::cell::RefCell;
use std::rc::Rc;

use crate::math::eqn::Eqns;
use crate::math::{Point2, Region, Region1, Region2, Scalar};
use std::fmt;

type EntityId = i64;

#[derive(Clone, Copy, Debug)]
pub struct Constrainable<T: Clone, TRegion: Region<T>> {
    id: EntityId,
    value: T,
    constraints: TRegion,
}

impl<T: Clone + fmt::Display, TRegion: Region<T> + fmt::Display> fmt::Display
    for Constrainable<T, TRegion>
{
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{{ {} ∈ {} }}", self.value, self.constraints)
    }
}

impl<T: Clone, TRegion: Region<T> + Clone> Constrainable<T, TRegion> {
    pub fn new(value: T, constraints: TRegion) -> Self {
        Self {
            id: 0,
            value,
            constraints,
        }
    }

    pub fn new_full(value: T) -> Self {
        Self::new(value, TRegion::full())
    }

    pub fn new_single(value: T) -> Self {
        Self::new(value.clone(), TRegion::singleton(value))
    }

    pub fn val(&self) -> &T {
        &self.value
    }

    pub fn resolve_with(&mut self, eqns: &Eqns) -> bool {
        let resolved_constraints = self.constraints.clone().substitute(eqns).simplify();
        if let Some(n) = resolved_constraints.nearest(&self.value) {
            self.value = n;
            true
        } else {
            false
        }
    }

    pub fn evaluate_with(&mut self, eqns: &Eqns) -> bool {
        self.constraints = self.constraints.clone().substitute(eqns).simplify();
        if let Some(n) = self.constraints.nearest(&self.value) {
            self.value = n;
            true
        } else {
            false
        }
    }

    pub fn constraints(&self) -> &TRegion {
        &self.constraints
    }

    pub fn reconstrain(&mut self, new_constraints: TRegion) -> bool {
        self.constraints = new_constraints;
        if let Some(n) = self.constraints.nearest(&self.value) {
            self.value = n;
            true
        } else {
            false
        }
    }
}

pub type CScalar = Constrainable<Scalar, Region1>;
pub type CPoint = Constrainable<Point2<Scalar>, Region2>;

pub type PointRef = Rc<RefCell<CPoint>>;

impl CPoint {
    pub fn into_ref(self) -> PointRef {
        Rc::new(RefCell::new(self))
    }
}

struct Line {
    p1: PointRef,
    p2: PointRef,
    len: CScalar,
    dir: CScalar,
}

// struct System {
//     points: Vec<Point>,
// }
