use std::borrow::{Borrow, Cow};
use std::collections::btree_map::Entry;
use std::collections::BTreeMap;
use std::fmt;

use super::expr::Expr;
use super::unknown::*;
use crate::math::expr::Expr::*;

#[derive(Clone, Debug, PartialEq)]
pub struct Eqn(pub Expr, pub Expr);

impl Unknowns for Eqn {
    fn unknowns(&self) -> UnknownSet {
        self.0
            .unknowns()
            .union(&self.1.unknowns())
            .cloned()
            .collect()
    }
    fn has_unknowns(&self) -> bool {
        self.0.has_unknowns() || self.1.has_unknowns()
    }
    fn has_unknown(&self, u: Unknown) -> bool {
        self.0.has_unknown(u) || self.1.has_unknown(u)
    }
}

fn ord_by_unkn(a: Expr, b: Expr, u: Unknown) -> Option<(Expr, Expr)> {
    if a.has_unknown(u) {
        Some((a, b))
    } else if b.has_unknown(u) {
        Some((b, a))
    } else {
        None
    }
}

impl Eqn {
    pub fn new(l: Expr, r: Expr) -> Self {
        Self(l, r)
    }

    pub fn simplify(self) -> Eqn {
        Eqn(self.0.simplify(), self.1.simplify())
    }

    pub fn substitute(self, eqns: &Eqns) -> Eqn {
        Self(self.0.substitute(eqns), self.1.substitute(eqns))
    }
}

pub fn solve_eqn(eqn: &Eqn, for_u: Unknown) -> Option<Expr> {
    use Expr::*;
    if !eqn.has_unknown(for_u) {
        return None;
    }
    trace!("solve: {}", eqn);
    let (mut l, mut r) = (eqn.0.clone().distribute().simplify(), eqn.1.clone().distribute().simplify());
    if l == r {
        return None
    };
    l = (l - r).distribute().simplify();
    r = Expr::Const(0.);
    loop {
        trace!("solve iter: {} == {}", l, r);
        let (new_l, new_r): (Expr, Expr) = match l {
            Unkn(u) => return if u == for_u { Some(r.simplify()) } else { None },
            Sum(es) => {
                let (us, not_us): (Vec<_>, Vec<_>) =
                    es.into_iter().partition(|e| e.has_unknown(for_u));
                if us.len() != 1 {
                    return None;
                }
                (
                    us.into_iter().next().unwrap(),
                    if not_us.len() == 0 {
                        r
                    } else {
                        Expr::new_minus(r, Sum(not_us))
                    },
                )
            }
            Product(es) => {
                let (us, not_us): (Vec<_>, Vec<_>) =
                    es.into_iter().partition(|e| e.has_unknown(for_u));
                if us.len() != 1 {
                    return None;
                }
                (
                    us.into_iter().next().unwrap(),
                    if not_us.len() == 0 {
                        r
                    } else {
                        Expr::new_div(r, Product(not_us))
                    },
                )
            }
            Neg(v) => (*v, Expr::new_neg(r)),
            Div(num, den) => {
                let (nu, du) = (num.has_unknown(for_u), den.has_unknown(for_u));
                match (nu, du) {
                    (true, false) => (*num, Expr::new_product(r, *den)),
                    (false, true) => (Expr::new_product(r, *den), *num),
                    (true, true) => return None, // TODO: simplify
                    (false, false) => return None,
                }
            }
            Const(_) => return None,
        };
        l = new_l.distribute().simplify();
        r = new_r.distribute().simplify();
    }
}

impl fmt::Display for Eqn {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{} == {}", self.0, self.1)
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Eqns<'a>(pub Cow<'a, [Eqn]>);

impl<'a> Unknowns for Eqns<'a> {
    fn unknowns(&self) -> UnknownSet {
        self.0.iter().flat_map(|eqn: &Eqn| eqn.unknowns()).collect()
    }
    fn has_unknowns(&self) -> bool {
        self.0.iter().any(|eqn: &Eqn| eqn.has_unknowns())
    }
    fn has_unknown(&self, u: Unknown) -> bool {
        self.0.iter().any(|eqn: &Eqn| eqn.has_unknown(u))
    }
}

impl<'a> fmt::Display for Eqns<'a> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{{ ")?;
        let mut first = true;
        for eq in (&self.0 as &Borrow<[Eqn]>).borrow() {
            if first {
                first = false;
            } else {
                write!(f, " && ")?;
            }
            write!(f, "{}", eq)?
        }
        write!(f, " }}")
    }
}

pub fn solve_eqns<'a>(eqns: &Eqns<'a>, for_vars: &[Unknown]) -> Option<Eqns<'static>> {
    let eqn_unknowns = eqns.unknowns();
    let has_all_vars = for_vars.iter().all(|u| eqn_unknowns.contains(u));
    if !has_all_vars {
        // return None;
    }
    let mut var_sols = BTreeMap::<Unknown, Vec<Expr>>::new();
    for u in &eqn_unknowns {
        for eq in (&eqns.0 as &Borrow<[Eqn]>).borrow() {
            let sol = solve_eqn(eq, *u);
            if let Some(sol) = sol {
                //println!("sol {0} -> {2} from {1}", u, eq, sol);
                match var_sols.entry(*u) {
                    Entry::Vacant(vac) => {
                        vac.insert(vec![sol]);
                    }
                    Entry::Occupied(mut occ) => {
                        if !occ.get().contains(&sol) {
                            occ.get_mut().push(sol);
                        };
                    }
                };
            } else {
                //println!("no sol for {0} in {1}", u, eq);
            }
        }
    }
    let mut new_eqns: Vec<Eqn> = Vec::new();
    let mut all_have_sols = true;
    for e in &var_sols {
        let mut has_sol = false;
        let mut last_expr = Unkn(*e.0);
        println!("{}", e.0);
        for s in e.1 {
            println!(" == {}", s);
            new_eqns.push(Eqn(last_expr, s.clone()));
            last_expr = s.clone();
            match s {
                Const(_) => { has_sol = true; }
                _ => {}
            };
        }
        if !has_sol {
            all_have_sols = false
        }
    }
    let mut new_eqns = Eqns(new_eqns.into());
    println!("new eqns: {}", new_eqns);
    if all_have_sols {
        return Some(new_eqns);
    }
    return solve_eqns(&new_eqns, for_vars);
    /*
    let mut more_eqns: Vec<Eqn> = Vec::new();
    let mut new_orig_eqns = eqns.clone();
    for eq in new_orig_eqns.0.to_mut() {
        let u = match eq.0 {
            Unkn(u) => u,
            _ => continue,
        };
        let my_new_eqns = Eqns(
            new_eqns
                .0
                .iter()
                .cloned()
                .filter(|e| !e.1.has_unknown(u))
                .collect::<Vec<_>>()
                .into(),
        );
        let r =
            eq.1.clone()
                .substitute(&my_new_eqns)
                .distribute()
                .simplify();
        more_eqns.push(Eqn(Unkn(u), r));
    }
    let mut more_eqns = Eqns(more_eqns.into());
    println!("more eqns: {}", more_eqns);
    */
    None
}

#[cfg(test)]
mod tests {
    use super::super::Scalar;
    use super::*;

    #[test]
    fn test_unknowns() {
        let u1 = Unknown(1);
        let u2 = Unknown(2);
        let u3 = Unknown(3);
        assert!(u1.has_unknowns());
        assert!(u2.has_unknowns());
        assert!(u1.has_unknown(u1));
        assert!(!u1.has_unknown(u2));
        assert!(u1.unknowns().contains(&u1));
        assert!(!u2.unknowns().contains(&u1));

        let e1 = Expr::new_minus(Expr::Unkn(u1), Expr::Unkn(u2));
        assert!(e1.has_unknowns());
        assert!(e1.has_unknown(u1));
        assert!(e1.has_unknown(u2));
        assert!(!e1.has_unknown(u3));
        assert!(e1.unknowns().len() == 2);
    }

    fn const_expr(e: Expr) -> Option<Scalar> {
        match e {
            Expr::Const(c) => Some(c),
            _ => None,
        }
    }

    #[test]
    fn test_solve_eqn() {
        use Expr::*;
        let _ = env_logger::try_init();
        let u1 = Unknown(1);
        let e1 = Unkn(u1);
        let e2 = Const(1.);

        let eqn = Eqn(e1.clone(), e2.clone());
        assert_eq!(solve_eqn(&eqn, u1), Some(Const(1.)));
        let eqn = Eqn(e2.clone(), e1.clone());
        assert_eq!(solve_eqn(&eqn, u1), Some(Const(1.)));

        let e3: Expr = Expr::from(1.) + 1. + 2.;
        let eqn = Eqn(e1.clone(), e3.clone());
        assert_eq!(solve_eqn(&eqn, u1), Some(Const(4.)));
        let e3 = Expr::from(1.) - 1.;
        let eqn = Eqn(e1.clone(), e3.clone());
        assert_eq!(solve_eqn(&eqn, u1), Some(Const(0.)));

        let e1 = Expr::from(2.) / (Expr::from(1.) - 4.);
        let e2 = Expr::new_minus(Const(1.), Unkn(u1));
        let eqn = Eqn(e1, e2);
        info!("eqn: {} => {}", eqn, eqn.clone().simplify());
        let e = solve_eqn(&eqn, u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), 5. / 3.));

        let e1 = Expr::from(2.) * (Expr::from(1.) - 4.);
        let e2 = (u1 * 2.) - u1;
        info!(
            "e1==e2: {}=={} => {}=={}",
            e1,
            e2,
            e1.clone().simplify(),
            e2.clone().simplify()
        );
        let eqn = Eqn(e1, e2);
        let e = solve_eqn(&eqn, u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), -6.));

        let e1 = Expr::from(2.) * (Expr::from(1.) - 4.);
        let e2 = ((u1 * 2.) + (u1 * u1)) / u1;
        info!(
            "{}=={} distrib=> {}=={}",
            e1,
            e2,
            e1.clone().distribute(),
            e2.clone().distribute()
        );
        info!(
            "{}=={} simplify=> {}=={}",
            e1,
            e2,
            e1.clone().distribute().simplify(),
            e2.clone().distribute().simplify()
        );
        let eqn = Eqn(e1, e2);
        let e = solve_eqn(&eqn, u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), -8.));

        let e1 = Expr::from(2.) * (Expr::from(1.) - (4.));
        let e2 = ((u1 * 2.) + (((u1 * u1) + u1) + (Expr::from(2.) - (1. + 1.)))) / u1;
        info!(
            "e1==e2: {}=={} => {}=={}",
            e1,
            e2,
            e1.clone().distribute().simplify(),
            e2.clone().distribute().simplify()
        );
        let eqn = Eqn(e1, e2);
        let e = solve_eqn(&eqn, u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), -9.));
    }

    #[test]
    fn test_solve_eqn2() {
        let _ = env_logger::try_init();

        let u1 = Unknown(1);
        let e1 = Expr::Unkn(u1);
        let e2 = (u1 * 2.) - 2.;
        info!(
            "e1==e2: {}=={} => {}=={}",
            e1,
            e2,
            e1.clone().distribute().simplify(),
            e2.clone().distribute().simplify(),
        );
        let eqn = Eqn(e1, e2);
        let e = solve_eqn(&eqn, u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), 2.));
    }

    #[test]
    fn test_solve_equations() {
        let _ = env_logger::try_init();

        use Expr::*;

        let x = Unknown(1);
        let y = Unknown(2);
        let t1 = Unknown(3);
        let t2 = Unknown(4);

        let eqns = Eqns(
            vec![
                Eqn::new(x.into(), t1 * 1.),
                Eqn::new(y.into(), t1 / 2.),
                Eqn::new(x.into(), Const(0.0) + t2 / 2.),
                Eqn::new(y.into(), t2 / 2.),
            ]
            .into(),
        );
        println!("eqns: {}", eqns);
        let sol = solve_eqns(&eqns, &[t1, t2]).unwrap();
        println!("sols: {}", sol);
    }

    #[test]
    fn test_solve_equations2() {
        let _ = env_logger::try_init();

        use Expr::*;

        let x = Unknown(1);
        let y = Unknown(2);
        let t1 = Unknown(3);
        let t2 = Unknown(4);

        let eqns = Eqns(
            vec![
                Eqn::new(x.into(), t1 * 1.),
                Eqn::new(y.into(), t1 / 2.),
                Eqn::new(x.into(), Const(1.0) + t2 / 1.),
                Eqn::new(y.into(), t2 / 2.),
            ]
            .into(),
        );
        println!("eqns: {}", eqns);
        let sol = solve_eqns(&eqns, &[t1, t2]).unwrap();
        println!("sols: {}", sol);
    }

}
