use std::borrow::Borrow;
use std::collections::BTreeMap;
use std::fmt;

use super::eqn::{Eqn, Eqns};
use super::unknown::{Unknown, UnknownSet, Unknowns};
use super::Scalar;
use std::f64::NAN;

#[derive(Clone, Debug, PartialEq)]
pub enum Expr {
    Unkn(Unknown),
    Const(Scalar),
    Sum(Exprs),
    Neg(Box<Expr>),
    Product(Exprs),
    Div(Box<Expr>, Box<Expr>),
}

pub type Exprs = Vec<Expr>;

impl Unknowns for Exprs {
    fn unknowns(&self) -> UnknownSet {
        self.iter().flat_map(|e: &Expr| e.unknowns()).collect()
    }
    fn has_unknowns(&self) -> bool {
        self.iter().any(|e: &Expr| e.has_unknowns())
    }
    fn has_unknown(&self, u: Unknown) -> bool {
        self.iter().any(|e: &Expr| e.has_unknown(u))
    }
}

fn write_separated_exprs(es: &[Expr], f: &mut fmt::Formatter, sep: &str) -> fmt::Result {
    write!(f, "(")?;
    let mut is_first = es.len() > 1;
    for e in es {
        if is_first {
            is_first = false;
        } else {
            write!(f, "{}", sep)?
        }
        write!(f, "{}", e)?
    }
    write!(f, ")")?;
    Ok(())
}

fn remove_common_terms(l: &mut Vec<Expr>, r: &mut Vec<Expr>) -> Vec<Expr> {
    let common: Vec<_> = l.drain_filter(|e| r.contains(e)).collect();
    common.iter().for_each(|e| {
        r.remove_item(e);
    });
    common
}

fn remove_term(terms: &mut Vec<Expr>, term: &Expr) -> Option<Expr> {
    terms.remove_item(term)
}

fn sum_fold(l: Expr, r: Expr) -> Expr {
    use Expr::*;
    match (l, r) {
        (Const(lc), Const(rc)) => Const(lc + rc),
        (Const(c), o) | (o, Const(c)) if relative_eq!(c, 0.) => o,
        (Product(mut l), Product(mut r)) => {
            let comm = remove_common_terms(&mut l, &mut r);
            if comm.is_empty() {
                Expr::new_sum(Product(l), Product(r))
            } else {
                Expr::new_product(Product(comm), Expr::new_sum(Product(l), Product(r)).collapse())
            }
        }
        (Product(mut l), r) | (r, Product(mut l)) => {
            let comm = remove_term(&mut l, &r);
            match comm {
                Some(_) => Expr::new_product(r, Expr::new_sum(Product(l), Const(1.))).collapse(),
                None => Expr::new_sum(Product(l), r),
            }
        }
        (Div(box ln, box ld), Div(box rn, box rd)) => {
            if ld == rd {
                Expr::new_div(Expr::new_sum(ln, rn), ld).collapse()
            } else {
                Expr::new_div(Expr::new_sum(Expr::new_product(ln, rd.clone()), Expr::new_product(rn, ld.clone())),
                    Expr::new_product(ld, rd)).collapse()
            }
        }
        (l, r) => Expr::new_sum(l, r),
    }
}

fn group_sum(es: Exprs) -> Exprs {
    use Expr::*;
    let mut common: BTreeMap<UnknownSet, Expr> = BTreeMap::new();
    for e in es {
        match e {
            Const(c) if relative_eq!(c, 0.) => {
                continue;
            }
            _ => (),
        };
        let unkns = e.unknowns();
        match common.get_mut(&unkns) {
            None => {
                common.insert(unkns, e);
            }
            Some(existing) => {
                match existing {
                    Sum(ref mut es) => {
                        // already failed at merging, so just add it to the list
                        es.push(e);
                    }
                    other => {
                        *other = sum_fold(other.clone(), e);
                    }
                };
            }
        };
    }
    for c in common.values() {
        trace!("group sum value: {}", c);
    }
    common
        .into_iter()
        .map(|(_, v)| v)
        .filter(|e| match e {
            Const(c) if relative_eq!(*c, 0.) => false,
            _ => true,
        })
        .collect()
}

fn product_fold(l: Expr, r: Expr) -> Expr {
    use Expr::*;
    match (l, r) {
        (Const(lc), Const(rc)) => Const(lc * rc),
        (Const(c), o) | (o, Const(c)) if relative_eq!(c, 1.) => o,
        (Const(c), _) | (_, Const(c)) if relative_eq!(c, 0.) => Const(0.),
        (Div(num, den), mul) | (mul, Div(num, den)) => {
            if mul == *den {
                *num
            } else {
                Expr::Div(Box::new(Expr::Product(vec![*num, mul])), den).simplify()
            }
        }
        (Product(mut ls), Product(mut rs)) => {
            ls.append(&mut rs);
            Product(ls)
        }
        (Product(mut ps), o) | (o, Product(mut ps)) => {
            ps.push(o);
            Product(ps)
        }
        (l, r) => Expr::new_product(l, r),
    }
}

fn group_product(es: Exprs) -> Exprs {
    use Expr::*;
    let es2 = es.clone();
    let mut consts: Option<Scalar> = None;
    let mut other = Exprs::new();
    for e in es {
        let unkns = e.unknowns();
        match e {
            Const(c) => match consts {
                None => consts = Some(c),
                Some(cs) => consts = Some(c * cs),
            },
            e => other.push(e),
        }
    }
    if let Some(cs) = consts {
        if relative_eq!(cs, 0.0) {
            other.clear();
            other.push(Const(0.0))
        } else if relative_ne!(cs, 1.0) {
            other.push(Const(cs))
        }
    };
    trace!("group product: {:?} => {:?}", es2, other);
    other
}

fn distribute_product_sums(mut es: Exprs) -> Expr {
    let es_pre = es.clone();
    use itertools::Itertools;
    use Expr::*;
    for e in &mut es {
        *e = e.clone().distribute();
    }
    let sums = es
        .drain_filter(|e| match e {
            Sum(_) => true,
            _ => false,
        })
        .map(|e| {
            trace!("sum in product: {}", e);
            match e.simplify() {
                Sum(es) => es,
                o => vec![o],
            }
        });
    let products: Vec<_> = sums.multi_cartesian_product().collect();
    if products.is_empty() {
        trace!("distribute_product_sums: no sums to distribute");
        return Product(es);
    }
    let sums = products
        .into_iter()
        .map(|mut prod| {
            prod.extend(es.clone());
            trace!("prod: {}", Product(prod.clone()));
            Product(prod)
        })
        .collect();
    let res = Sum(sums);
    trace!("distribute_product_sums: {} => {}", Product(es_pre), res);
    res
}

impl Unknowns for Expr {
    fn unknowns(&self) -> UnknownSet {
        use Expr::*;
        match self {
            Unkn(u) => u.unknowns(),
            Const(_) => UnknownSet::default(),
            Sum(es) | Product(es) => es.unknowns(),
            Div(l, r) => l.unknowns().union(&r.unknowns()).cloned().collect(),
            Neg(e) => e.unknowns(),
        }
    }
    fn has_unknowns(&self) -> bool {
        use Expr::*;
        match self {
            Unkn(u) => u.has_unknowns(),
            Const(_) => false,
            Sum(es) | Product(es) => es.has_unknowns(),
            Div(l, r) => l.has_unknowns() || r.has_unknowns(),
            Neg(e) => e.has_unknowns(),
        }
    }
    fn has_unknown(&self, u: Unknown) -> bool {
        use Expr::*;
        match self {
            Unkn(u1) => u1.has_unknown(u),
            Const(_) => false,
            Sum(es) | Product(es) => es.has_unknown(u),
            Div(l, r) => l.has_unknown(u) || r.has_unknown(u),
            Neg(e) => e.has_unknown(u),
        }
    }
}

impl Expr {
    pub fn new_sum(e1: Expr, e2: Expr) -> Expr {
        Expr::Sum(vec![e1, e2])
    }
    pub fn new_product(e1: Expr, e2: Expr) -> Expr {
        Expr::Product(vec![e1, e2])
    }
    pub fn new_neg(e1: Expr) -> Expr {
        Expr::Neg(Box::new(e1))
    }
    pub fn new_div(num: Expr, den: Expr) -> Expr {
        Expr::Div(Box::new(num), Box::new(den))
    }
    pub fn new_minus(e1: Expr, e2: Expr) -> Expr {
        Expr::Sum(vec![e1, Expr::new_neg(e2)])
    }
    pub fn new_inv(den: Expr) -> Expr {
        Expr::new_div(Expr::Const(1.), den)
    }

    pub fn is_zero(&self) -> bool {
        use Expr::*;
        match self.clone().simplify() {
            Const(c) => relative_eq!(c, 0.),
            _ => false,
        }
    }

    pub fn is_one(&self) -> bool {
        use Expr::*;
        match self.clone().simplify() {
            Const(c) => relative_eq!(c, 1.),
            _ => false,
        }
    }

    pub fn substitute<'a>(self, eqns: &Eqns<'a>) -> Expr {
        use Expr::*;
        for eqn in (&eqns.0 as &Borrow<[Eqn]>).borrow() {
            if self == eqn.0 {
                return eqn.1.clone();
            }
        }
        match self {
            Sum(mut es) => {
                for e in &mut es {
                    *e = e.clone().substitute(eqns);
                }
                Sum(es)
            }
            Product(mut es) => {
                for e in &mut es {
                    *e = e.clone().substitute(eqns);
                }
                Product(es)
            }
            Neg(mut e) => {
                *e = e.substitute(eqns);
                Neg(e)
            }
            Div(mut num, mut den) => {
                *num = num.substitute(eqns);
                *den = den.substitute(eqns);
                Div(num, den)
            }
            other => other,
        }
    }

    pub fn collapse(self) -> Expr {
        use Expr::*;
        match self {
            Sum(es) => {
                let mut new_es: Vec<_> = es
                    .into_iter()
                    .map(|e| e.collapse())
                    .flat_map(|e| match e {
                        Sum(more_es) => more_es,
                        other => vec![other],
                    })
                    .collect();

                let consts = new_es.drain_filter(|e| match e {
                    Const(_) => true,
                    _ => false
                }).fold(Const(0.), |l, r| match (l, r) {
                    (Const(lc), Const(rc)) => Const(lc + rc),
                    _ => unreachable!(),
                });
                new_es.push(consts);

                match new_es.len() {
                    0 => Const(0.),                          // none
                    1 => new_es.into_iter().next().unwrap(), // one
                    _ => Sum(new_es),                        // many
                }
            }
            Product(es) => {
                let new_es: Vec<_> = es
                    .into_iter()
                    .map(|e| e.collapse())
                    .flat_map(|e| match e {
                        Product(more_es) => more_es,
                        other => vec![other],
                    })
                    .collect();
                match new_es.len() {
                    0 => Const(1.),                          // none
                    1 => new_es.into_iter().next().unwrap(), // one
                    _ => {
                        if new_es.iter().any(|e| e.is_zero()) {
                            Const(0.)
                        } else {
                            Product(new_es)
                        }
                    },                    // many
                }
            }
            Neg(mut v) => {
                *v = v.collapse();
                match v {
                    box Const(c) => Const(-c),
                    box Neg(v) => *v,
                    box Product(mut es) => {
                        es.push(Const(-1.));
                        Product(es).collapse()
                    }
                    e => Product(vec![Const(-1.), *e]).collapse(),
                }
            }
            Div(mut num, mut den) => {
                *num = num.collapse();
                *den = den.collapse();
                match (num, den) {
                    (box Const(num), box Const(den)) => Const(num / den),
                    (box Const(num), den) => {
                        if relative_eq!(num, 0.) {
                            Const(0.)
                        } else {
                            Div(Box::new(Const(num)), den)
                        }
                    }
                    (num, box Const(den)) => {
                        if relative_eq!(den, 1.) {
                            *num
                        } else {
                            Expr::new_product(*num, Const(1. / den)).collapse()
                        }
                    }
                    (num, box Div(dennum, denden)) => {
                        Div(Box::new(Product(vec![*num, *denden])), dennum).collapse()
                    }
                    (box Product(mut es), box den) => match es.remove_item(&den) {
                        Some(_) => Product(es),
                        None => Expr::new_div(Product(es), den),
                    },
                    (num, den) => {
                        if num == den {
                            Expr::Const(1.)
                        } else {
                            Div(num, den)
                        }
                    }
                }
            }
            e => e,
        }
    }

    pub fn simplify(self) -> Expr {
        use Expr::*;
        match self {
            Sum(es) => {
                let pre_new_es = es.clone();
                let mut new_es: Vec<_> = es
                    .into_iter()
                    .map(|e| e.simplify())
                    .flat_map(|e| match e {
                        Sum(more_es) => more_es,
                        other => vec![other],
                    })
                    .collect();
                new_es = group_sum(new_es);
                trace!(
                    "simplify sum {} => {}",
                    Sum(pre_new_es),
                    Sum(new_es.clone())
                );

                match new_es.len() {
                    0 => Const(0.),                          // none
                    1 => new_es.into_iter().next().unwrap(), // one
                    _ => Sum(new_es),                        // many
                }
            }
            Product(es) => {
                let pre_new_es = es.clone();
                let new_es: Vec<_> = es
                    .into_iter()
                    .map(|e| e.simplify())
                    .flat_map(|e| match e {
                        Product(more_es) => more_es,
                        other => vec![other],
                    })
                    .collect();
                let new_es = group_product(new_es);
                trace!(
                    "simplify product {} => {}",
                    Product(pre_new_es),
                    Product(new_es.clone())
                );
                match new_es.len() {
                    0 => Const(1.),                          // none
                    1 => new_es.into_iter().next().unwrap(), // one
                    _ => Product(new_es),                    // many
                }
            }
            Neg(mut v) => {
                *v = v.simplify();
                trace!("simplify neg {}", Neg(v.clone()));
                match v {
                    box Const(c) => Const(-c),
                    box Neg(v) => *v,
                    box Product(mut es) => {
                        es.push(Const(-1.));
                        Product(es).simplify()
                    }
                    e => Product(vec![Const(-1.), *e]).simplify(),
                }
            }
            Div(mut num, mut den) => {
                *num = num.simplify();
                *den = den.simplify();
                trace!("simplify div {}", Div(num.clone(), den.clone()));
                match (num, den) {
                    (box Const(num), box Const(den)) => Const(num / den),
                    (num, box Const(den)) => {
                        if relative_eq!(den, 1.) {
                            *num
                        } else {
                            Expr::new_product(*num, Const(1. / den)).simplify()
                        }
                    }
                    (num, box Div(dennum, denden)) => {
                        Div(Box::new(Product(vec![*num, *denden])), dennum).simplify()
                    }
                    (box Product(mut es), box den) => match es.remove_item(&den) {
                        Some(_) => Product(es).simplify(),
                        None => Expr::new_div(Product(es), den),
                    },
                    (num, den) => {
                        if num == den {
                            Expr::Const(1.)
                        } else {
                            Div(num, den)
                        }
                    }
                }
            }
            Const(c) => if c.is_nan() {
                println!("NAN!");
                Const(c)
            } else { Const(c) },
            e => e,
        }
    }

    pub fn distribute(self) -> Expr {
        use Expr::*;
        match self {
            Sum(mut es) => {
                let es_pre = es.clone();
                for e in &mut es {
                    *e = e.clone().distribute();
                }
                let res = Sum(es);
                trace!("distribute sum {} => {}", Sum(es_pre), res);
                res
            }
            Product(es) => distribute_product_sums(es),
            Div(num, den) => match (num, den) {
                (box Sum(es), box den) => Sum(es
                    .into_iter()
                    .map(|e| Expr::new_div(e, den.clone()).distribute())
                    .collect()),
                (num, den) => Div(num, den),
            },
            Neg(v) => match v {
                box Const(c) => Const(-c),
                box Sum(mut es) => {
                    for e in &mut es {
                        *e = Expr::new_neg(e.clone()).distribute();
                    }
                    Sum(es)
                }
                box Product(mut es) => {
                    for e in &mut es {
                        *e = e.clone().distribute();
                    }
                    es.push(Const(-1.));
                    Product(es)
                }
                box Neg(v) => v.distribute(),
                box Div(mut num, mut den) => {
                    *num = Neg(num.clone());
                    Div(num, den).distribute()
                }
                e => Neg(Box::new(e.distribute())),
            },
            e => e,
        }
    }
}

impl fmt::Display for Expr {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        use Expr::*;
        match self {
            Unkn(u) => write!(f, "{}", u),
            Const(c) => write!(f, "{}", c),
            Sum(es) => write_separated_exprs(es, f, " + "),
            Product(es) => write_separated_exprs(es, f, " * "),
            Div(num, den) => write!(f, "({} / {})", num, den),
            Neg(e) => write!(f, "-({})", e),
        }
    }
}
