use crate::entity::{CPoint as PointEntity, PointRef};
use crate::math::{GenericRegion, Line2, Point2, Region1, Region2, Rot2, Scalar, Value, Vec2};

#[derive(Clone, Copy, Debug, PartialEq)]
pub enum ResolveResult {
    Underconstrained,
    Constrained,
    Overconstrained,
}

impl ResolveResult {
    pub fn from_r2(r: &Region2) -> ResolveResult {
        use Region2::*;
        match r {
            Empty => ResolveResult::Overconstrained,
            Singleton(_) => ResolveResult::Constrained,
            _ => ResolveResult::Constrained,
        }
    }
}

pub trait Relation {
    fn resolve(&self) -> ResolveResult;
}

pub struct Coincident {
    pub p1: PointRef,
    pub p2: PointRef,
}

impl Relation for Coincident {
    fn resolve(&self) -> ResolveResult {
        let (mut p1, mut p2) = (self.p1.borrow_mut(), self.p2.borrow_mut());
        let r = {
            p1.constraints()
                .clone()
                .intersect(p2.constraints().clone())
                .simplify()
        };
        p1.reconstrain(r.clone());
        p2.reconstrain(r.clone());
        ResolveResult::from_r2(&r)
    }
}

pub struct PointAngle {
    pub p1: PointRef,
    pub p2: PointRef,
    pub angle: Rot2,
}

impl PointAngle {
    pub fn new(p1: PointRef, p2: PointRef, angle: Rot2) -> Self {
        Self { p1, p2, angle }
    }

    pub fn new_horizontal(p1: PointRef, p2: PointRef) -> Self {
        Self::new(
            p1,
            p2,
            Rot2::from_cos_sin_unchecked(1., 0.),
        )
    }

    pub fn new_vertical(p1: PointRef, p2: PointRef) -> Self {
        Self::new(
            p1,
            p2,
            Rot2::from_cos_sin_unchecked(0., 1.),
        )
    }
}

impl Relation for PointAngle {
    fn resolve(&self) -> ResolveResult {
        use Region2::*;
        let (mut p1, mut p2) = (self.p1.borrow_mut(), self.p2.borrow_mut());
        let constrain_line = |p1: &Point2<Value>, p2: &mut PointEntity| {
            let line = Region2::Line(Line2::new(p1.clone(), self.angle, Region1::Full));
            trace!(
                "PointAngle line: {}, p2 constraint: {}",
                line,
                p2.constraints()
            );
            let new_constraint = p2.constraints().clone().intersection(line).simplify();
            trace!("PointAngle new_constraint: {}", new_constraint);
            p2.reconstrain(new_constraint);
            ResolveResult::from_r2(p2.constraints())
        };
        match (&mut p1.constraints(), &mut p2.constraints()) {
            (Empty, _) | (_, Empty) => ResolveResult::Overconstrained,
            (Singleton(p1), Singleton(p2)) => {
                // if the angle p1 and p2 form is parallel to self.angle, the result
                // will have a y component of 0
                let r = self.angle.conj() * (p2.clone() - p1.clone());
                trace!("angle.cos: {}", r.x);
                // if relative_eq!(r.y, 0.) {
                ResolveResult::Constrained
                // } else {
                // ResolveResult::Overconstrained
                // }
            }
            (Singleton(p), _) => constrain_line(p, &mut *p2),
            (_, Singleton(p)) => constrain_line(p, &mut *p1),
            _ => ResolveResult::Underconstrained,
        }
    }
}

pub struct AlignedDistance {
    pub p1: PointRef,
    pub p2: PointRef,
    pub angle: Rot2,
    pub distance: Scalar,
}

impl AlignedDistance {
    pub fn new(p1: PointRef, p2: PointRef, angle: Rot2, distance: Scalar) -> Self {
        Self {
            p1,
            p2,
            angle,
            distance,
        }
    }

    pub fn new_vertical(p1: PointRef, p2: PointRef, distance: Scalar) -> Self {
        Self::new(p1, p2, Rot2::up(), distance)
    }

    pub fn new_horizontal(p1: PointRef, p2: PointRef, distance: Scalar) -> Self {
        Self::new(p1, p2, Rot2::right(), distance)
    }
}

impl Relation for AlignedDistance {
    fn resolve(&self) -> ResolveResult {
        use Region2::*;
        let (mut p1, mut p2) = (self.p1.borrow_mut(), self.p2.borrow_mut());
        let constrain_line = |p1: Point2<Value>, p2: &mut PointEntity| {
            let angle = self.angle + Rot2::up();
            let line = Region2::Line(Line2::new(p1.clone(), angle, Region1::Full)).simplify();
            trace!(
                "AlignedDistance line: {}, p2 constraint: {}",
                line,
                p2.constraints()
            );
            let new_constraint = p2.constraints().clone().intersection(line).simplify();
            trace!("AlignedDistance new_constraint: {}", new_constraint);
            p2.reconstrain(new_constraint);
            ResolveResult::from_r2(p2.constraints())
        };
        let offset: Vec2<Scalar> = self.angle * self.distance;
        match (&mut p1.constraints(), &mut p2.constraints()) {
            (Empty, _) | (_, Empty) => ResolveResult::Overconstrained,
            (Singleton(p1), Singleton(p2)) => {
                let r = p2.clone() - p1.clone();
                let d = self.angle.dot(r);
                // if relative_eq!(d, self.distance) {
                ResolveResult::Constrained
                // } else {
                // ResolveResult::Overconstrained
                // }
            }
            (Singleton(pos), _) => constrain_line(pos.clone() + offset, &mut *p2),
            (_, Singleton(pos)) => constrain_line(pos.clone() - offset, &mut *p1),
            _ => ResolveResult::Underconstrained,
        }
    }
}

pub struct Midpoint {
    pub p1: PointRef,
    pub p2: PointRef,
    pub mid: PointRef,
}
